"""
Parsing functions used throughout portscanner
"""
from contextlib import suppress
from functools import partial

from portscanner.utils import trycast


__all__ = [
    "port_range",
]


def port_range(port_ranges: str, quiet: bool = True):
    """
    Parse port ranges such as:
        80
        80,443
        21-23,80,443,8000-8080
    """
    valid_ports = set()
    for port_range in port_ranges.split(","):
        try:
            if "-" not in port_range:
                if (port := trycast(int, port_range)) is not None:
                    if 0 < port < 65536:
                        valid_ports.add(port)
            elif port_range.count("-") == 1:
                start, stop = map(
                    lambda s: trycast(int, s.strip()), port_range.split("-")
                )
                if start is None or stop is None:
                    raise ValueError
                start, stop = (stop, start) if start > stop else (start, stop)
                for i in range(max((0, start)), min((stop, 65535)) + 1):
                    valid_ports.add(i)
            else:
                raise ValueError
        except ValueError:
            if not quiet:
                print("Invalid Port Range:", port_range)
    return sorted(valid_ports)



"""
WorkPool implementation to allow for limited execution of an unbounded number of tasks
"""
from abc import ABC, abstractmethod, abstractproperty
from contextlib import suppress
from typing import Coroutine, Callable, Iterable
import asyncio

__all__ = [
    "MxWorkPoolBase",
    "MxWorkPool",
]


class MxWorkPoolBase(ABC):
    """
    Provides async multi-worker functionality to a class
    """

    @abstractmethod
    def __init__(self, worker_count: int):
        """Initialize the work pool"""

    @abstractproperty
    def worker_available(self) -> bool:
        """Return True if workers are available, False if semahore is locked"""

    @abstractproperty
    def worker_sem(self) -> asyncio.Semaphore:
        """Read-only handle to the worker semaphore used"""

    @abstractproperty
    def worker_count(self) -> int:
        """Number of workers"""

    @abstractmethod
    async def worker_run(self, coro: Coroutine) -> asyncio.Future:
        """Run a single corotuine in the pool and return its corresponding future"""

    @abstractmethod
    async def worker_run_many(self, coros: Coroutine) -> asyncio.Future:
        """Run a single corotuine in the pool and return its corresponding future"""


class MxWorkPool(MxWorkPoolBase):
    """Workpool implementation"""

    def __init__(self, worker_count: int):
        """Set worker count and create the shared semaphore"""
        self._worker_count = worker_count
        self.__sem = asyncio.Semaphore(worker_count)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if hasattr(self, "_loop"):
            return self._loop
        return asyncio.get_event_loop()

    @property
    def worker_available(self) -> bool:
        return not self.__sem.locked()

    @property
    def worker_sem(self) -> asyncio.Semaphore:
        return self.__sem

    @property
    def worker_count(self) -> int:
        return self._worker_count

    def worker_run(self, coro: Coroutine, *callbacks) -> asyncio.Future:
        async def worker():
            # with suppress(asyncio.CancelledError):
            async with self.worker_sem:
                return await coro

        fut = self.loop.create_task(worker())
        for callback in callbacks:
            fut.add_done_callback(callback)
        return fut

    async def worker_run_many(self, coros: Iterable[Coroutine], timeout: float = 1.0):
        futures = set()
        init = False

        async def check(t):
            nonlocal futures
            if futures:
                done, futures = await asyncio.wait(
                    futures, timeout=t, return_when=asyncio.FIRST_COMPLETED
                )
                for task in done:
                    if not task.cancelled() and not task.exception():
                        yield task.result()

        wait = 0
        try:
            while futures or not init:
                init = True
                async for item in check(wait):
                    yield item
                if len(futures) < self.worker_count:
                    wait = 0
                    with suppress(StopIteration):
                        futures.add(self.worker_run(next(coros)))
                        continue
                wait = timeout
        except asyncio.CancelledError:
            for future in futures:
                future.cancel()



"""
Utilities and helper functions/classes for PortScanner
"""
from collections import namedtuple
from functools import partial, wraps
from itertools import product
from portscanner.types import IPNetwork
from typing import Collection, Iterable, Text
import asyncio


__all__ = [
    "splat",
    "cleanup",
    "bind",
    "trycast",
    "trycast_many",
    "flatten",
]


def splat(func):
    """
    Wrapper for a fuction that will simply accept a tuple and splat into the function
    """

    @wraps(func)
    def splatter(args: tuple, **kwargs):
        return func(*args, **kwargs)

    return splatter


def cleanup(loop: asyncio.AbstractEventLoop):
    """
    Safe cleanup method to destroy all pending tasks
    """
    while (pending := asyncio.all_tasks(loop)) :
        for task in pending:
            task.cancel()
        loop.run_until_complete(asyncio.wait(pending))
    loop.close()


class bind(partial):
    """
    An improved version of partial which accepts Ellipsis (...) as a placeholder
    """

    def __call__(self, *args, **keywords):
        keywords = {**self.keywords, **keywords}
        iargs = iter(args)
        args = (next(iargs) if arg is ... else arg for arg in self.args)
        return self.func(*args, *iargs, **keywords)


def trycast(new_type, value, default=None):
    """
    Attempt to cast `value` as `new_type` or `default` if conversion fails
    """
    try:
        default = new_type(value)
    finally:
        return default


def trycast_many(new_type, value_iter, default=None, flat: bool = False):
    """
    Generator to yield only valid cast items from an iterables
    """
    for value in flatten(value_iter) if flat else value_iter:
        if (new_value := trycast(new_type, value, default)) is not default:
            yield new_value


def flatten(*values):
    """
    Flatten one or several iterables into one generator
    """
    for value in values:
        if not isinstance(value, Text) and isinstance(value, Iterable):
            for val in value:
                yield from flatten(val)
        else:
            yield value


def flattener(**iterables):
    """
    Nested loop through named iterables. The names provided of the iterables
    are the name names provided in the output, but each `yield` will be
    equivalent to nested
    """
    names, iters = zip(*iterables.items())
    Value = namedtuple("Value", names)

    for values in product(*map(flatten, iters)):
        yield Value(*values)



"""
DNS Resolver mixin for PortScanner
"""

from abc import ABC, abstractmethod, abstractproperty
from contextlib import suppress
from ipaddress import ip_address, ip_network
from typing import Union, Optional, List, Collection, get_args
import asyncio

from aiodns import DNSResolver
from aiodns.error import DNSError
from portscanner.types import IPAddress, IPNetwork, IPAny, IPTypes

__all__ = [
    "MxResolverBase",
    "MxResolver",
]


class MxResolverBase:
    @abstractproperty
    def resolver(self) -> DNSResolver:
        """
        Return the resolver used for the mixin
        Should be located at `self._resolver` by the child class
        """

    @abstractproperty
    def timeout(self) -> float:
        """
        Return the resolver timeout used for the mixin
        Should be located at `self._timeout` by the child class
        """

    @abstractmethod
    async def resolve_all(self, host: IPAny, qtype: str) -> List[IPNetwork]:
        """
        Resolve a host into all of its IP addresses.
        Upon error or NXDOMAIN, return an empty list
        """

    @abstractmethod
    async def resolve(self, host: IPAny, qtype: str) -> Optional[IPNetwork]:
        """
        Resolve a host into the first IP address returned by `resolve_all`
        Upon error or NXDOMAIN, return None
        """


class MxResolver(MxResolverBase):
    DEFAULT_TIMEOUT = 1.0

    @property
    def resolver(self) -> DNSResolver:
        return self._resolver

    @resolver.setter
    def resolver(self, resolver: DNSResolver):
        self._resolver = resolver

    @property
    def timeout(self) -> float:
        return getattr(self, "_timeout", self.DEFAULT_TIMEOUT)

    @timeout.setter
    def timeout(self, timeout: float):
        self._timeout = timeout

    async def resolve_all(self, host: IPAny, qtype: str = "A") -> List[IPNetwork]:
        """
        Return a list of valid IPv4/IPv6 addresses
        """
        if isinstance(host, get_args(IPTypes)):
            return [host]
        with suppress(ValueError):
            return [ip_network(host, strict=False)]
        if isinstance(qtype, str):
            qtypes = [qtype]
        elif isinstance(qtype, Collection):
            qtypes = list(qtype)
        else:
            raise ValueError(f"Invalid Query Type '{qtype}'")
        tasks = set()
        for qtype in self._Q if qtypes is None else qtypes:
            if (qtype := qtype.upper()) not in self._Q:
                raise ValueError(f"Supported query types: {', '.join(self._Q)}")
            tasks.add(asyncio.wait_for(self.resolver.query(host, qtype), self.timeout))
        with suppress(TimeoutError, asyncio.TimeoutError, DNSError):
            results, _ = await asyncio.wait(tasks, timeout=self.timeout)
            return list(
                ip_network(r.host, strict=False)
                for res in results
                for r in res.result()
                if not res.cancelled() or res.exception()
            )
        return []

    async def resolve(self, host: IPAny, qtype: str = "A") -> Optional[IPTypes]:
        if isinstance(host, get_args(IPTypes)):
            return host
        ips = await self.resolve_all(host, qtype)
        return ips[0] if ips else None



"""
Loop acquisition mixin for PortScanner
"""

from abc import ABC, abstractstaticmethod, abstractproperty
from contextlib import suppress
from typing import Optional
import asyncio

__all__ = [
    "MXLoopBase",
    "MXLoop",
]


class MxLoopBase(ABC):
    @abstractproperty
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        Get the stored event loop or return one from the environment
        Should be stored in `self._loop` in the child class
        """

    @abstractstaticmethod
    def _get_loop() -> asyncio.AbstractEventLoop:
        """
        Get the environment loop. It is up to the implementation
        if you'd like to raise an exception or create and set a new loop
        """


class MxLoop(MxLoopBase):
    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        self._loop = loop or self._get_loop()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop or self._get_loop()

    @staticmethod
    def _get_loop() -> Optional[asyncio.AbstractEventLoop]:
        with suppress(RuntimeError):
            return getattr(
                asyncio,
                "get_running_loop",
                asyncio.get_event_loop,
            )()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop



"""
Custom types used within PortScanner
"""
from dataclasses import dataclass
from enum import IntEnum, auto
from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network
from typing import Optional, Text, Union


__all__ = [
    "IPAddress",
    "IPNetwork",
    "IPTypes",
    "IPAny",
    "ScanState",
    "ScanInfo",
]


# Type aliases
IPAddress = Union[IPv4Address, IPv6Address]
IPNetwork = Union[IPv4Network, IPv6Network]
IPTypes = Union[IPAddress, IPNetwork]
IPAny = Union[IPTypes, Text]


class ScanState(IntEnum):
    """
    Possible values for the state of a particular host/port
    """

    OPEN = auto()  # Port is open and a reply was received
    CLOSED = auto()  # Connection was actively refused
    TIMEOUT = auto()  # Port did not respond within the desired time frame
    UNKNOWN = auto()  # Unknown error occurred (unlikely)


@dataclass
class ScanInfo:
    """
    Returned info from an individual port scan
    """

    host: str
    port: int
    state: ScanState
    banner: Optional[str] = None



"""
Implementation of an asynchronous port scanner
"""

from codecs import decode
from contextlib import suppress
from dataclasses import dataclass
from enum import IntEnum, auto
from functools import partial
from ipaddress import ip_address
from socket import AF_INET, AF_INET6, AF_UNSPEC
from typing import Optional, List, Union, Sequence, get_args, Iterable, Collection
import asyncio
import sys
import time

from aiodns import DNSResolver
from aiodns.error import DNSError

from portscanner.mixins import MxLoop, MxResolver, MxWorkPool
from portscanner.types import IPAddress, IPNetwork, IPTypes, IPAny, ScanInfo, ScanState
from portscanner.utils import trycast, trycast_many, flattener, flatten


class PortScanner(MxLoop, MxResolver, MxWorkPool):
    """
    Implementation of the port scanner
    """

    _NS = ("1.1.1.1", "1.0.0.1", "8.8.8.8", "8.8.4.4")
    _Q = ("A", "AAAA")

    def __init__(
        self,
        workers: int = 10,
        timeout: float = 3.0,
        banner_buffer: Optional[int] = None,
        loop: asyncio.AbstractEventLoop = None,
        resolver: DNSResolver = None,
    ):
        MxLoop.__init__(self, loop)
        MxWorkPool.__init__(self, workers)
        self._timeout = timeout
        self._banner_buffer = banner_buffer
        self._resolver = resolver or DNSResolver(self._NS, loop=self.loop)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return True

    def _run(self, coro):
        return asyncio.wait_for(coro, timeout=self._timeout)

    async def scan(
        self,
        hosts: Collection[IPAny],
        ports: Sequence[int],
        open: bool = False,
        qtype: str = "A",
        all: bool = True,
        verbose: bool = False,
    ):
        resolved_hosts = []

        # Resolve all hosts first before scanning is initiated
        resolver = self.resolve_all if all else self.resolve
        tasks = (
            resolver(host, qtype=qt) for qt in flatten(qtype) for host in iter(hosts)
        )
        if verbose:
            start = time.time()
        async for resolved_host in self.worker_run_many(tasks):
            if resolved_host:
                if isinstance(resolved_host, List):
                    resolved_hosts += resolved_host
                else:
                    resolved_hosts.append(resolved_host)

        if verbose:
            print(
                f"Resolution of {len(hosts)} hosts took {time.time()-start:.3f} seconds"
            )
            total_hosts = sum(rh.num_addresses for rh in resolved_hosts)
            print("Scanning", len(ports), "ports on", total_hosts, "hosts")

        data_gen = flattener(host=resolved_hosts, port=ports)
        tasks = (self._scan_port(*data) for data in data_gen)

        if verbose:
            start = time.time()

        async for scan_info in self.worker_run_many(tasks):
            if not open or scan_info.state == ScanState.OPEN:
                yield scan_info

        if verbose:
            print(f"Scan executed in {time.time() - start:.3f} seconds")

    async def worker(self, *args):
        import random

        r = random.randint(1, 10)
        await asyncio.sleep(r / 10.0)
        return r

    async def scan_port(
        self, host: Union[IPAddress, str], port: int, qtype: str = "A"
    ) -> ScanInfo:
        return await self.worker_run(self._scan_port(host, port, qtype))

    async def _scan_port(
        self, host: Union[IPAddress, str], port: int, qtype: str = "A"
    ) -> ScanInfo:
        try:
            if (host := await self.resolve(host)) is None:
                raise ValueError(f"Invalid host '{host}'")
            if isinstance(host, get_args(IPNetwork)):
                host = host[0]

            family = {
                4: AF_INET,
                6: AF_INET6,
            }.get(host.version, AF_UNSPEC)

            # Set initial values
            state, banner, reader, writer = ScanState.UNKNOWN, None, None, None
            fut = asyncio.open_connection(host=str(host), port=port, family=family)
            reader, writer = await self._run(fut)
            state = ScanState.OPEN
            if self._banner_buffer:
                writer.write(b"")
                task = self._run(reader.read(self._banner_buffer))
                table = str.maketrans("", "", "\r\n")
                decoder = partial(decode, encoding="utf-8", errors="ignore")
                banner = decoder(await task).translate(table)
        except asyncio.CancelledError:
            pass
        except asyncio.TimeoutError:
            if state == ScanState.UNKNOWN:
                state = ScanState.TIMEOUT
        except (ConnectionRefusedError, OSError) as e:
            if state == ScanState.UNKNOWN:
                state = ScanState.CLOSED
        except Exception as e:
            print(f"Error ({type(e)}):\n{e}", file=sys.stderr)
        finally:
            if writer:
                writer.close()
                await writer.wait_closed()
            return ScanInfo(host=host, port=port, state=state, banner=banner)



from .resolver import MxResolverBase, MxResolver
from .loop import MxLoopBase, MxLoop
from .workpool import MxWorkPoolBase, MxWorkPool

__all__ = [
    "MxResolverBase", "MxResolver",
    "MxLoopBase", "MxLoop",
    "MxWorkPoolBase", "MxWorkPool",
]


"""
Colors used for output, particularly in the run module
"""

from colorama import init, Fore, Style


__all__ = [
    "Clr",
]


class Clr:
    """3-Letter Color Constants"""

    init = init

    RST = Style.RESET_ALL
    BRT = Style.BRIGHT

    GRN = BRT + Fore.GREEN
    BLU = BRT + Fore.BLUE
    CYN = BRT + Fore.CYAN
    YLW = BRT + Fore.YELLOW
    PRP = BRT + Fore.MAGENTA
    RED = BRT + Fore.RED
    WHT = BRT + Fore.WHITE
